/*
 * Copyright (c) 1995, 1996, 1997, 1998 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#define __NO_VERSION__

#include "xfs_locl.h"
#include "xfs_dev.h"
#include "xfs_cache.h"
#include "xfs_dirent.h"
#include "xfs_syscalls.h"
#include <linux/binfmts.h>
#include <linux/file.h>

RCSID("$Id: xfs_inodeops.c,v 1.30 1998/06/08 22:46:00 map Exp $");

static int
#ifndef LINUX2_1
xfs_lookup (struct inode * dir, const char * name, int len,
	    struct inode ** result);
#else
xfs_lookup (struct inode *inode, struct dentry *dentry);
#endif /* LINUX2_1 */

static int
#ifndef LINUX2_1
xfs_fsync(struct inode *, struct file *);
#else
xfs_fsync(struct file *file, struct dentry *dentry);
#endif /* LINUX2_1 */

static int
xfs_open(struct inode *i, struct file *f);

static int
xfs_permission(struct inode *inode, int mode);

#ifndef LINUX2_1
static int
xfs_read_file(struct inode *inode, struct file *file, char *buf, int count);
#else
static ssize_t
xfs_read_file(struct file *file, char *buf, size_t count, loff_t *ppos);
#endif /* LINUX2_1 */

#ifndef LINUX2_1
static int
xfs_write_file(struct inode *inode, struct file *file, const char *buf, int count);
#else
static ssize_t
xfs_write_file(struct file *file, const char *buf, size_t count, loff_t *ppos);
#endif /* LINUX2_1 */

static int
#ifndef LINUX2_1
xfs_create (struct inode * dir, const char * name, int len, int mode, struct inode ** result);
#else
xfs_create (struct inode * dir, struct dentry *dentry, int mode);
#endif /* LINUX2_1 */

static int
#ifndef LINUX2_1
xfs_mkdir(struct inode * dir, const char * name, int len, int mode);
#else
xfs_mkdir(struct inode * dir, struct dentry *dentry, int mode);
#endif /* LINUX2_1 */

static int
#ifndef LINUX2_1
xfs_rmdir(struct inode * dir, const char * name, int len);
#else
xfs_rmdir(struct inode * dir, struct dentry *dentry);
#endif /* LINUX2_1 */

static int
#ifndef LINUX2_1
xfs_readdir(struct inode * inode, struct file * file,
	    void * dirent, filldir_t filldir);
#else
xfs_readdir(struct file * file, void * dirent, filldir_t filldir);
#endif /* LINUX2_1 */

#ifndef LINUX2_1
int
xfs_rename (struct inode * old_dir, const char * old_name, int old_len,
	     struct inode * new_dir, const char * new_name, int new_len,
	     int must_be_dir);

#else
static int
xfs_rename (struct inode * old_dir, struct dentry *old_dentry,
	     struct inode * new_dir, struct dentry *new_dentry);
#endif /* LINUX2_1 */

static int
#ifndef LINUX2_1
xfs_unlink (struct inode * dir, const char * name, int len);
#else
xfs_unlink (struct inode * dir, struct dentry *dentry);
#endif /* LINUX2_1 */

static int
#ifndef LINUX2_1
xfs_readlink(struct inode *inode, char *buffer, int buflen);
#else
xfs_readlink(struct dentry *dentry, char *buffer, int buflen);
#endif /* LINUX2_1 */

#ifndef LINUX2_1
static int
xfs_follow_link(struct inode * dir, struct inode * inode,
		int flag, int mode, struct inode ** res_inode);
#else
static struct dentry *
xfs_follow_link(struct dentry *, struct dentry *);
#endif /* LINUX2_1 */

#ifndef LINUX2_1
static void
#else
static int
#endif /* LINUX2_1 */
xfs_release_file (struct inode *inode, struct file *file);

static struct file_operations xfs_file_operations = {
        NULL,                   /* lseek */
        xfs_read_file,          /* read */
        xfs_write_file,         /* write */
        NULL,                   /* readdir */
        NULL,                   /* select */
        NULL,                   /* ioctl */
        NULL,                   /* mmap */
        xfs_open,               /* open */
        xfs_release_file,       /* release */
        xfs_fsync,              /* fsync */
	NULL,			/* fasync */
	NULL,			/* check_media_change */
	NULL			/* revalidate */
};

struct inode_operations xfs_file_inode_operations = {
        &xfs_file_operations,   /* default file operations */
        NULL,                   /* create */
        NULL,                   /* lookup */
        NULL,                   /* link */
        NULL,                   /* unlink */
        NULL,                   /* symlink */
        NULL,                   /* mkdir */
        NULL,                   /* rmdir */
        NULL,                   /* mknod */
        NULL,                   /* rename */
        NULL,                   /* readlink */
        NULL,                   /* follow_link */
        generic_readpage,       /* readpage */
        NULL,                   /* writepage */
        NULL,                   /* bmap */
        NULL,                   /* truncate (not yet) */
        xfs_permission,         /* permission */
        NULL                    /* smap */
};

static struct file_operations xfs_dead_operations = {
        NULL,                   /* lseek */
        NULL,                   /* read */
        NULL,                   /* write */
        NULL,                   /* readdir */
        NULL,                   /* select */
        NULL,                   /* ioctl */
        NULL,                   /* mmap */
        NULL,                   /* open */
        NULL,                   /* release */
        NULL,                   /* fsync */
	NULL,			/* fasync */
	NULL,			/* check_media_change */
	NULL			/* revalidate */
};

struct inode_operations xfs_dead_inode_operations = {
        &xfs_dead_operations,   /* default file operations */
        NULL,                   /* create */
        NULL,                   /* lookup */
        NULL,                   /* link */
        NULL,                   /* unlink */
        NULL,                   /* symlink */
        NULL,                   /* mkdir */
        NULL,                   /* rmdir */
        NULL,                   /* mknod */
        NULL,                   /* rename */
        NULL,                   /* readlink */
        NULL,                   /* follow_link */
        NULL,                   /* readpage */
        NULL,                   /* writepage */
        NULL,                   /* bmap */
        NULL,                   /* truncate */
        NULL,                   /* permission */
        NULL                    /* smap */
};

static struct file_operations xfs_dir_operations = {
        NULL,                   /* lseek */
        NULL,                   /* read */
        NULL,                   /* write */
        xfs_readdir,            /* readdir */
        NULL,                   /* select */
        NULL,                   /* ioctl */
        NULL,                   /* mmap */
        NULL,                   /* open */
        NULL,                   /* release */
        NULL,                   /* fsync */
	NULL,			/* fasync */
	NULL,			/* check_media_change */
	NULL			/* revalidate */
};

struct inode_operations xfs_dir_inode_operations = {
        &xfs_dir_operations,   /* default file operations */
        xfs_create,             /* create */
        xfs_lookup,             /* lookup */
        NULL,                   /* link (not yet) */
        xfs_unlink,             /* unlink */
        NULL,                   /* symlink (not yet) */
        xfs_mkdir,              /* mkdir */
        xfs_rmdir,              /* rmdir */
        NULL,                   /* mknod */
        xfs_rename,             /* rename */
        NULL,                   /* readlink */
        NULL,                   /* follow_link */
        NULL,                   /* readpage */
        NULL,                   /* writepage */
        NULL,                   /* bmap */
        NULL,                   /* truncate */
        xfs_permission,         /* permission */
        NULL                    /* smap */
};

static struct file_operations xfs_link_operations = {
        NULL,                   /* lseek */
        NULL,                   /* read */
        NULL,                   /* write */
        NULL,                   /* readdir */
        NULL,                   /* select */
        NULL,                   /* ioctl */
        NULL,                   /* mmap */
        NULL,                   /* open */
        NULL,                   /* release */
        NULL,                   /* fsync */
	NULL,			/* fasync */
	NULL,			/* check_media_change */
	NULL			/* revalidate */
};

struct inode_operations xfs_link_inode_operations = {
        &xfs_link_operations,   /* default file operations */
        NULL,                   /* create */
        NULL,                   /* lookup */
        NULL,                   /* link */
        NULL,                   /* unlink */
        NULL,                   /* symlink */
        NULL,                   /* mkdir */
        NULL,                   /* rmdir */
        NULL,                   /* mknod */
        NULL,                   /* rename */
        xfs_readlink,           /* readlink */
#ifndef LINUX2_1
        xfs_follow_link,        /* follow_link */
#else
	NULL,
#endif /* LINUX2_1 */
        NULL,                   /* readpage */
        NULL,                   /* writepage */
        NULL,                   /* bmap */
        NULL,                   /* truncate */
        NULL,                   /* permission */
        NULL                    /* smap */
};

#ifndef LINUX2_1
static int
xfs_lookup (struct inode * dir, const char * name, int len,
	    struct inode ** result)
{
    struct xfs_message_getnode msg;
    struct xfs *xfsp;
    int error = 0;
    
    struct xfs_node *d;
    struct xfs_node *x;
    
    XFSDEB(XDEBVNOPS, ("xfs_lookup: name: %s(%d) ", name, len));
    
    if (dir == NULL) {
        XFSDEB(XDEBVNOPS, ("ENOENT\n"));
        return -ENOENT;
    }
    if (!S_ISDIR(dir->i_mode)) {
        iput(dir);
        XFSDEB(XDEBVNOPS, ("ENOTDIR\n"));
        return -ENOTDIR;
    }
    xfsp = XFS_FROM_VNODE(dir);
    d = VNODE_TO_XNODE(dir);
    if (d == NULL || xfsp == NULL) {
        iput(dir);
        XFSDEB(XDEBVNOPS, ("ENODEV\n"));
        return -ENODEV;
    }

    *result = NULL;
    do {
#ifdef notdef_but_correct_and_not_ported
        error = xfs_access(dvp, VEXEC, cred);
        if (error != 0)
            goto done;
#endif

        if ((x = xfs_cache_lookup(VNODE_TO_XNODE(dir), name, len)) == NULL) {
            msg.header.opcode = XFS_MSG_GETNODE;

            msg.cred.uid = current->uid;
            msg.cred.pag = xfs_get_pag();
            msg.parent_handle = d->handle;
            
            memcpy(msg.name, name, len);
            msg.name[len]='\0';
            
            error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
            
            if (error == 0)
                error = ((struct xfs_message_wakeup *) &msg)->error;
        } else {
            *result=iget(dir->i_sb,(ino_t) x);
            goto done;
        }
    } while (error == 0);

done:
    if (error)
        XFSDEB(XDEBVNOPS, ("error %d", error));
    XFSDEB(XDEBVNOPS, ("\n"));
    iput(dir);
    return -error;
}
#else
static int
xfs_lookup (struct inode *dir, struct dentry *dentry)
{
    struct xfs_message_getnode msg;
    struct xfs *xfsp;
    int error = 0;
    
    struct xfs_node *d;
    
    XFSDEB(XDEBVNOPS, ("xfs_lookup: name: %s(%d) ", dentry->d_name.name, dentry->d_name.len));
    
    if (dir == NULL) {
        XFSDEB(XDEBVNOPS, ("ENOENT\n"));
        return -ENOENT;
    }
    xfsp = XFS_FROM_VNODE(dir);
    d = VNODE_TO_XNODE(dir);
    if (d == NULL || xfsp == NULL) {
        XFSDEB(XDEBVNOPS, ("ENODEV\n"));
        return -ENODEV;
    }

    do {
        struct dentry   *res_dentry;

#ifdef notdef_but_correct_and_not_ported
        error = xfs_access(dvp, VEXEC, cred);
        if (error != 0)
            goto done;
#endif

        if ((res_dentry = d_lookup(list_entry(dir->i_dentry.next,
                                              struct dentry, d_alias),
                                   &dentry->d_name)) != NULL) {
            if (IS_ERR(res_dentry)) {
                XFSDEB(XDEBVNOPS, ("xfs_lookup: d_lookup error: %ld \n",
                                   PTR_ERR(res_dentry)));
                error = PTR_ERR(res_dentry);
            } else if (res_dentry == NULL) {
                printk("XFS Panic: d_lookup returned NULL\n");
            } else {
                /* memcpy(dentry, res_dentry, sizeof(*dentry)); */
                d_add(dentry, res_dentry->d_inode);
                d_drop(res_dentry);
            }
            goto done;
        }

        msg.header.opcode = XFS_MSG_GETNODE;
        
        msg.cred.uid = current->uid;
        msg.cred.pag = xfs_get_pag();
        msg.parent_handle = d->handle;
        
        memcpy(msg.name, dentry->d_name.name, dentry->d_name.len);
        msg.name[dentry->d_name.len]='\0';
        
        error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
        
        if (error == 0)
            error = ((struct xfs_message_wakeup *) &msg)->error;
    } while (error == 0);

done:
    if (error == ENOENT) {
        XFSDEB(XDEBVNOPS, ("xfs_lookup: adding negative cache\n"));
        d_add (dentry, NULL);
        error = 0;
    }
    if (error)
        XFSDEB(XDEBVNOPS, ("error %d", error));
    XFSDEB(XDEBVNOPS, ("\n"));
    return -error;
}
#endif /* LINUX2_1 */


static int
xfs_open_valid(struct inode *vp, u_int tok)
{
  struct xfs *xfsp = XFS_FROM_VNODE(vp);
  struct xfs_node *xn = VNODE_TO_XNODE(vp);
  int error = 0;

  XFSDEB(XDEBVFOPS, ("xfs_open_valid\n"));
  
  do {
    if (!XFS_TOKEN_GOT(xn, tok))
      {
	struct xfs_message_open msg;
	msg.header.opcode = XFS_MSG_OPEN;
	msg.cred.uid = current->uid;
	msg.cred.pag = xfs_get_pag();
	msg.handle = xn->handle;
	msg.tokens = tok;
	error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
	if (error == 0)
	  error = ((struct xfs_message_wakeup *) &msg)->error;
      }
    else
      {
	goto done;
      }
  } while (error == 0);

 done:
  return -error;
}

static int
xfs_open(struct inode *i, struct file *f)
{
  XFSDEB(XDEBVNOPS, ("xfs_open f->f_mode: %d\n",f->f_mode));

  if (f->f_mode & FMODE_WRITE)
    return xfs_open_valid(i, XFS_OPEN_NW);
  else
    return xfs_open_valid(i, XFS_OPEN_NR);
}

static int
do_fsync(struct xfs *xfsp, struct xfs_node *xn, u_int flag)
{
    int error;
    struct xfs_message_putdata msg;

    msg.header.opcode = XFS_MSG_PUTDATA;
    msg.cred.uid = current->uid;
    msg.cred.pag = xfs_get_pag();
    msg.handle = xn->handle;
    msg.flag = flag;

    error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
    if (error == 0)
	error = ((struct xfs_message_wakeup *) & msg)->error;

    if (error == 0)
	xn->flags &= ~XFS_DATA_DIRTY;

    return error;
}

static int
#ifndef LINUX2_1
xfs_fsync(struct inode *inode, struct file *file)
#else
xfs_fsync(struct file *file, struct dentry *dentry)
#endif /* LINUX2_1 */
{
#ifdef LINUX2_1
    struct inode *inode = dentry->d_inode;
#endif /* LINUX2_1 */
    struct xfs *xfsp = XFS_FROM_VNODE(inode);
    struct xfs_node *xn = VNODE_TO_XNODE(inode);
    int error = 0;

    XFSDEB(XDEBVNOPS, ("xfs_fsync: 0x%p\n", inode));

    if (xn == NULL)
	return 0;

    if (xn->flags & XFS_DATA_DIRTY)
	error = do_fsync(xfsp, xn, XFS_DATA_DIRTY);
    return error;
}

static int
xfs_attr_valid(struct inode * vp, u_int tok)
{
    struct xfs *xfsp = XFS_FROM_VNODE(vp);
    struct xfs_node *xn = VNODE_TO_XNODE(vp);
    int error = 0;

    pag_t pag = xfs_get_pag();

    do {
        if (!XFS_TOKEN_GOT(xn, tok)) {
            struct xfs_message_getattr msg;

            msg.header.opcode = XFS_MSG_GETATTR;
            msg.cred.uid = current->uid;
            msg.cred.pag = pag;
            msg.handle = xn->handle;
            error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
            if (error == 0)
                error = ((struct xfs_message_wakeup *) & msg)->error;
        } else {
            goto done;
        }
    } while (error == 0);

done:
    return error;
}

static int
xfs_fetch_rights(struct inode *vp)
{
    struct xfs *xfsp = XFS_FROM_VNODE(vp);
    struct xfs_node *xn = VNODE_TO_XNODE(vp);
    int error = 0;

    pag_t pag = xfs_get_pag();

    struct xfs_message_getattr msg;

    msg.header.opcode = XFS_MSG_GETATTR;
    msg.cred.uid = current->uid;
    msg.cred.pag = pag;
    msg.handle = xn->handle;
    error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
    if (error == 0)
	error = ((struct xfs_message_wakeup *) & msg)->error;

    return (error);
}

static int
xfs_rights_valid(struct inode * vp, pag_t pag)
{
    struct xfs *xfsp = XFS_FROM_VNODE(vp);
    struct xfs_node *xn = VNODE_TO_XNODE(vp);
    int error = 0;
    printk("pag: %d\n", pag);

    do {
        if (!xfs_has_pag(xn, pag))
        {
            struct xfs_message_getattr msg;

            msg.header.opcode = XFS_MSG_GETATTR;
            msg.cred.uid = current->uid;
            msg.cred.pag = pag;
            msg.handle = xn->handle;
            error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
            if (error == 0)
                error = ((struct xfs_message_wakeup *) & msg)->error;
        }
        else {
            goto done;
        }
    } while (error == 0);

done:
    return error;
}


static int
check_rights (u_char rights, int mode)
{
    int error = 0;

    if (mode & MAY_READ)
	if ((rights & XFS_RIGHT_R) == 0)
	    error = EACCES;
    if (mode & MAY_WRITE)
	if ((rights & XFS_RIGHT_W) == 0)
	    error = EACCES;
    if (mode & MAY_EXEC)
	if ((rights & XFS_RIGHT_X) == 0)
	    error = EACCES;
    return error;
}


static int
xfs_permission(struct inode *inode, int mode)
{
    int error = 0;
    pag_t pag = xfs_get_pag();
    
    XFSDEB(XDEBVNOPS, ("xfs_access mode = 0%o\n", mode));

    if (VNODE_TO_XNODE(inode) == NULL) {
       printk("XFS Panic! xnode == NULL in xfs_permission\n");
       printk("   inode = %p i_ino: 0x%lx\n", inode, inode->i_ino);
       return -EACCES;
    }

    error = xfs_attr_valid(inode, XFS_ATTR_R);
    if (error == 0) {
	struct xfs_node *xn = VNODE_TO_XNODE(inode);
	int i;

	error = check_rights (xn->anonrights, mode);
	
	if (error == 0)
	    goto done;

	XFSDEB(XDEBVNOPS, ("xfs_access anonaccess failed\n"));

	xfs_rights_valid(inode, pag); /* ignore error */
	
	error = EACCES;
	
	for (i = 0; i < MAXRIGHTS; i++)
	    if (xn->id[i] == pag) {
		error = check_rights (xn->rights[i], mode);
		break;
	    }
    }

done:
    XFSDEB(XDEBVNOPS, ("xfs_access(0%o) = %d\n", mode, error));
    return -error;
}

static int
xfs_data_valid(struct inode *vp, u_int tok)
{
  struct xfs *xfsp = XFS_FROM_VNODE(vp);
  struct xfs_node *xn = VNODE_TO_XNODE(vp);
  int error = 0;
  
  do {
    if (!XFS_TOKEN_GOT(xn, tok))
      {
	struct xfs_message_getdata msg;
	msg.header.opcode = XFS_MSG_GETDATA;
	msg.cred.uid = current->uid;
	msg.cred.pag = xfs_get_pag();
	msg.handle = xn->handle;
	msg.tokens = tok;
	error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
	if (error == 0)
	  error = ((struct xfs_message_wakeup *) &msg)->error;
      }
    else
      {
	goto done;
      }
  } while (error == 0);

 done:
  return -error;
}


#ifndef LINUX2_1
static int
xfs_read_file(struct inode *inode, struct file *file, char *buf, int count)
{
  int error = 0;

  XFSDEB(XDEBVNOPS, ("xfs_read_file file->f_pos: %d count: %d\n",(int) file->f_pos,count));

  error = xfs_data_valid(inode, XFS_DATA_R);

  if (error == 0) {
    struct inode *t = DATA_FROM_VNODE(inode);
    t->i_count++;
    if (t->i_op &&
        t->i_op->default_file_ops &&
        t->i_op->default_file_ops->read)
      error = t->i_op->default_file_ops->read(t, file, buf, count);
    else {      
        printk("XFS Panic! read not defined in cache file system\n");
        error = -EINVAL;
    }
    iput(t);
  }
  return error;
}
#else
static int
xfs_read_file(struct file *file, char *buf, size_t count, loff_t *ppos)
{
    int error = 0;
    struct inode *inode = file->f_dentry->d_inode;
    
    XFSDEB(XDEBVNOPS, ("xfs_read_file file->f_pos: %d count: %d\n",(int) file->f_pos,count));
    
    error = xfs_data_valid(inode, XFS_DATA_R);
    
    if (error == 0) {
        struct dentry *t = DATA_FROM_VNODE(inode);
        struct dentry cache_dentry;
        struct file cache_file;
        
        cache_file.f_pos = file->f_pos;
        cache_file.f_mode = file->f_mode;
        cache_file.f_flags = file->f_flags;
        cache_file.f_count = file->f_count;
        cache_file.f_owner = file->f_owner;
        cache_file.f_op = t->d_inode->i_op->default_file_ops;
        cache_file.f_dentry = &cache_dentry;
        cache_file.f_dentry->d_inode = t->d_inode;

        if (cache_file.f_op == NULL ||
            cache_file.f_op->read == NULL) {
            printk("XFS Panic! write not defined in cache file system\n");
            return -EINVAL;
        }

        error = cache_file.f_op->read(&cache_file,
                                      buf,
                                      count,
                                      &(cache_file.f_pos));

        file->f_pos = cache_file.f_pos;
        inode->i_size = t->d_inode->i_size;
    }
    return error;
}
#endif /* LINUX2_1 */


#ifndef LINUX2_1
static int xfs_write_file(struct inode *inode, struct file *file, const char *buf, int count)
{
    int error = 0;
    
    XFSDEB(XDEBVNOPS, ("xfs_write_file file->f_pos: %d count: %d\n",(int) file->f_pos,count));
    
    error = xfs_data_valid(inode, XFS_DATA_W);
    
    if (error == 0) {
        struct inode *t = DATA_FROM_VNODE(inode);
        t->i_count++;
        if (t->i_op &&
            t->i_op->default_file_ops &&
            t->i_op->default_file_ops->read) {
            XFSDEB(XDEBVNOPS, ("xfs_write_file t:%p",t));
            error = t->i_op->default_file_ops->write(t, file, buf, count);
            XFSDEB(XDEBVNOPS, ("xfs_write_file error: %d\n",error));
            VNODE_TO_XNODE(inode)->flags |= XFS_DATA_DIRTY;
            inode->i_size = t->i_size;
            inode->i_mtime = t->i_mtime;
        } else {        
            printk("XFS Panic! write not defined in cache file system\n");
            error = -EINVAL;
        }
        iput(t);
    }
    return error;
}
#else
static ssize_t
xfs_write_file(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
    int error = 0;
    struct inode *inode = file->f_dentry->d_inode;
    
    XFSDEB(XDEBVNOPS, ("xfs_write_file file->f_pos: %d count: %d\n",(int) file->f_pos,count));
    
    error = xfs_data_valid(inode, XFS_DATA_W);
    
    if (error == 0) {
        struct dentry *t = DATA_FROM_VNODE(inode);
        struct dentry cache_dentry;
        struct file cache_file;
        
        cache_file.f_pos = file->f_pos;
        cache_file.f_mode = file->f_mode;
        cache_file.f_flags = file->f_flags;
        cache_file.f_count = file->f_count;
        cache_file.f_owner = file->f_owner;
        cache_file.f_op = t->d_inode->i_op->default_file_ops;
        cache_file.f_dentry = &cache_dentry;
        cache_file.f_dentry->d_inode = t->d_inode;

        if (cache_file.f_op == NULL ||
            cache_file.f_op->write == NULL) {
            printk("XFS Panic! write not defined in cache file system\n");
            return -EINVAL;
        }
        
        down(&t->d_inode->i_sem);
        error = cache_file.f_op->write(&cache_file,
                                       buf,
                                       count,
                                       &(cache_file.f_pos));
        up(&t->d_inode->i_sem);

        file->f_pos = cache_file.f_pos;
        inode->i_size = t->d_inode->i_size;
        inode->i_mtime = t->d_inode->i_mtime;
        VNODE_TO_XNODE(inode)->flags |= XFS_DATA_DIRTY;
    }
    return error;
}
#endif /* LINUX2_1 */


static int
#ifndef LINUX2_1
xfs_create (struct inode * dir, const char * name, int len, int mode, struct inode ** result)
#else
xfs_create (struct inode * dir, struct dentry *dentry, int mode)
#endif /* LINUX2_1 */
{
    struct xfs *xfsp;
    struct xfs_node *xn;
    int error = 0;
    struct xfs_message_create msg;
#ifdef LINUX2_1
    struct dentry *res_dentry;
#endif /* LINUX2_1 */

#ifndef LINUX2_1
    *result = NULL;
#endif /* LINUX2_1 */
    if (!dir)
	return -ENOENT;

    xfsp = XFS_FROM_VNODE(dir);
    xn = VNODE_TO_XNODE(dir);

#ifndef LINUX2_1
    XFSDEB(XDEBVNOPS, ("xfs_create: (%s, %d)\n",
		       name, len));
#else
    XFSDEB(XDEBVNOPS, ("xfs_create: (%s, %d)\n",
		       dentry->d_name.name, dentry->d_name.len));
#endif /* LINUX2_1 */

    msg.header.opcode = XFS_MSG_CREATE;
    msg.parent_handle = xn->handle;
#ifndef LINUX2_1
    strncpy(msg.name, name, 256);
#else
    strncpy(msg.name, dentry->d_name.name, 256);
#endif /* LINUX2_1 */
    XA_CLEAR(&msg.attr);
    XA_SET_MODE(&msg.attr, mode);
    XA_SET_UID(&msg.attr, current->uid);
    XA_SET_GID(&msg.attr, current->gid);
    XA_SET_TYPE(&msg.attr, XFS_FILE_REG);
    msg.mode = 0;		/* XXX */
    msg.cred.uid = current->uid;
    msg.cred.pag = xfs_get_pag();
    error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
    if (error == 0)
	error = ((struct xfs_message_wakeup *) & msg)->error;

#ifndef LINUX2_1
    if (error == 0)
	error = xfs_lookup(dir, name, len, result);
#else
    if (error == 0 &&
	(res_dentry = d_lookup(list_entry(dir->i_dentry.next,
					  struct dentry, d_alias),
			       &dentry->d_name)) != NULL) {
	if (IS_ERR(res_dentry)) {
	    XFSDEB(XDEBVNOPS, ("xfs_lookup: d_lookup error: %ld \n",
				   PTR_ERR(res_dentry)));
	    error = PTR_ERR(res_dentry);
	} else if (res_dentry == NULL) {
	    printk("XFS Panic: d_lookup returned NULL\n");
	} else {
	    /* memcpy(dentry, res_dentry, sizeof(*dentry)); */
	    d_instantiate(dentry, res_dentry->d_inode);
	    d_drop(res_dentry);
	    }
    }
#endif /* LINUX2_1 */

    return error;
    
}

int
#ifndef LINUX2_1
xfs_unlink (struct inode * dir, const char * name, int len)
#else
xfs_unlink (struct inode * dir, struct dentry *dentry)
#endif /* LINUX2_1 */
{
    struct xfs_message_remove msg;
    struct xfs *xfsp  = XFS_FROM_VNODE(dir);
    struct xfs_node *xn = VNODE_TO_XNODE(dir);
#ifndef LINUX2_1
    struct xfs_node *xn_file;
#endif /* LINUX2_1 */
    int error;
    
#ifndef LINUX2_1
    XFSDEB(XDEBVNOPS, ("xfs_remove: (%s, %d)\n", name, len));
#else
    XFSDEB(XDEBVNOPS, ("xfs_remove: (%s, %d)\n",
		       dentry->d_name.name,
		       dentry->d_name.len));
#endif /* LINUX2_1 */
    
    msg.header.opcode = XFS_MSG_REMOVE;
    msg.parent_handle = xn->handle;
#ifndef LINUX2_1
    strncpy(msg.name, name, 256);
#else
    strncpy(msg.name, dentry->d_name.name, 256);
#endif /* LINUX2_1 */
    msg.cred.uid = current->uid;
    msg.cred.pag = xfs_get_pag();
    error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
    if (error == 0)
        error = ((struct xfs_message_wakeup *) &msg)->error;

#ifndef LINUX2_1
    xn_file = xfs_cache_lookup(xn, name, len);
    if (xn_file != NULL)
	xfs_cache_delete(xn_file);
    iput(dir);
#else
    if (error == 0)
	d_delete(dentry);
#endif /* LINUX2_1 */

    return error;
}

int
#ifndef LINUX2_1
xfs_rename (struct inode * old_dir, const char * old_name, int old_len,
	     struct inode * new_dir, const char * new_name, int new_len,
	     int must_be_dir)
#else
xfs_rename (struct inode * old_dir, struct dentry *old_dentry,
	     struct inode * new_dir, struct dentry *new_dentry)
#endif /* LINUX2_1 */
{
    struct xfs *xfsp = XFS_FROM_VNODE(old_dir);
    struct xfs_message_rename msg;
#ifndef LINUX2_1
    struct xfs_node *xn_file;
#endif /* LINUX2_1 */
    int error;

    if (old_dir->i_sb != new_dir->i_sb)
	return -EXDEV;
    
    msg.header.opcode = XFS_MSG_RENAME;
    msg.old_parent_handle = VNODE_TO_XNODE(old_dir)->handle;
#ifndef LINUX2_1
    strncpy(msg.old_name, old_name, 256);
#else
    strncpy(msg.old_name, old_dentry->d_name.name, 256);
#endif /* LINUX2_1 */
    msg.new_parent_handle = VNODE_TO_XNODE(new_dir)->handle;
#ifndef LINUX2_1
    strncpy(msg.new_name, new_name, 256);
#else
    strncpy(msg.new_name, new_dentry->d_name.name, 256);
#endif /* LINUX2_1 */
    msg.cred.uid = current->uid;
    msg.cred.pag = xfs_get_pag();
    error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
    if (error == 0)
	error = ((struct xfs_message_wakeup *) &msg)->error;

#ifndef LINUX2_1
    xn_file = xfs_cache_lookup(VNODE_TO_XNODE(old_dir), old_name, old_len);
    xfs_cache_delete(xn_file);
    iput(old_dir);
    iput(new_dir);
#else
    if (error == 0)
	d_move(old_dentry, new_dentry);
#endif /* LINUX2_1 */

    return -error;
}

static int
#ifndef LINUX2_1
xfs_mkdir(struct inode * dir, const char * name, int len, int mode)
#else
xfs_mkdir(struct inode * dir, struct dentry *dentry, int mode)
#endif /* LINUX2_1 */
{
    struct xfs *xfsp;
    struct xfs_node *xn;
#ifdef LINUX2_1
    struct dentry *res_dentry;
#endif /* LINUX2_1 */
    int error = 0;

#ifndef LINUX2_1
    XFSDEB(XDEBVNOPS, ("xfs_mkdir name:%s\n",name));
#else
    XFSDEB(XDEBVNOPS, ("xfs_mkdir name:%s\n", dentry->d_name.name));
#endif /* LINUX2_1 */

    if (!dir)
	return -ENOENT;
#ifndef LINUX2_1
    if (len > 255) {
	iput (dir);
#else
    if (dentry->d_name.len > 255) {
#endif /* LINUX2_1 */
	return -ENAMETOOLONG;
    }

    xfsp = XFS_FROM_VNODE(dir);
    xn = VNODE_TO_XNODE(dir);

    {
	struct xfs_message_mkdir msg;

	msg.header.opcode = XFS_MSG_MKDIR;
	msg.parent_handle = xn->handle;
#ifndef LINUX2_1
	strncpy(msg.name, name, 256);
#else
	strncpy(msg.name, dentry->d_name.name, 256);
#endif /* LINUX2_1 */
	XA_CLEAR(&msg.attr);
	XA_SET_MODE(&msg.attr, mode);
	XA_SET_UID(&msg.attr, current->uid);
	XA_SET_GID(&msg.attr, current->gid);
	XA_SET_TYPE(&msg.attr, XFS_FILE_DIR);

	msg.cred.uid = current->uid;
	msg.cred.pag = xfs_get_pag();
	error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
	if (error == 0)
	    error = ((struct xfs_message_wakeup *) & msg)->error;
#ifdef LINUX2_1
	if (error == 0 &&
	    (res_dentry = d_lookup(list_entry(dir->i_dentry.next,
					      struct dentry, d_alias),
				   &dentry->d_name)) != NULL) {
	    if (IS_ERR(res_dentry)) {
		XFSDEB(XDEBVNOPS, ("xfs_lookup: d_lookup error: %ld \n",
				   PTR_ERR(res_dentry)));
		error = PTR_ERR(res_dentry);
	    } else if (res_dentry == NULL) {
		printk("XFS Panic: d_lookup returned NULL\n");
	    } else {
		/* memcpy(dentry, res_dentry, sizeof(*dentry)); */
		d_instantiate(dentry, res_dentry->d_inode);
		d_drop(res_dentry);
	    }
	}
#endif /* LINUX2_1 */
    }

#ifndef LINUX2_1
    iput(dir);
#endif /* LINUX2_1 */

    return -error;
}

static int
#ifndef LINUX2_1
xfs_rmdir(struct inode * dir, const char * name, int len)
#else
xfs_rmdir(struct inode * dir, struct dentry *dentry)
#endif /* LINUX2_1 */
{
    struct xfs *xfsp;
    struct xfs_node *xn;
#ifndef LINUX2_1
    struct xfs_node *xn_file;
#endif /* LINUX2_1 */
    struct xfs_message_rmdir msg;
    int error;

#ifndef LINUX2_1
    XFSDEB(XDEBVNOPS, ("xfs_rmdir: (%s, %d\n", name, len));
#else
    XFSDEB(XDEBVNOPS, ("xfs_rmdir: (%s, %d\n",
		       dentry->d_name.name,
		       dentry->d_name.len));
#endif /* LINUX2_1 */

    if (!dir)
	return -ENOENT;
#ifndef LINUX2_1
    if (len > 255) {
	iput (dir);
#else
    if (dentry->d_name.len > 255) {
#endif /* LINUX2_1 */
	return -ENAMETOOLONG;
    }

    xfsp = XFS_FROM_VNODE(dir);
    xn = VNODE_TO_XNODE(dir);

    msg.header.opcode = XFS_MSG_RMDIR;
    msg.parent_handle = xn->handle;
#ifndef LINUX2_1
    strncpy(msg.name, name, 256);
#else
    strncpy(msg.name, dentry->d_name.name, 256);
#endif /* LINUX2_1 */
    msg.cred.uid = current->uid;
    msg.cred.pag = xfs_get_pag();
    error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
    if (error == 0)
	error = ((struct xfs_message_wakeup *) &msg)->error;

    if (error == 0) {
#ifndef LINUX2_1
	xn_file = xfs_cache_lookup(xn, name, len);
	if (xn_file != NULL)
	    xfs_cache_delete(xn_file);
#else
	d_delete(dentry);
#endif /* LINUX2_1 */
    }

#ifndef LINUX2_1
    iput(dir);
#endif /* LINUX2_1 */

    return -error;
}

#ifndef LINUX2_1
static int xfs_readdir(struct inode * inode, struct file * file,
		       void * dirent, filldir_t filldir)
#else
static int
xfs_readdir(struct file * file, void * dirent, filldir_t filldir)
#endif /* LINUX2_1 */
{
    int error = 0;
    int filldir_error;
    off_t offset, begin_offset;
#ifndef LINUX2_1
    struct xfs_node *xn_file;
#else
    struct inode *inode = file->f_dentry->d_inode;
#endif /* LINUX2_1 */
    int fileno;
    char *buf;
#ifndef LINUX2_1
    long oldfs;
#endif /* LINUX2_1 */

    XFSDEB(XDEBREADDIR, ("xfs_readdir\n"));
    
    error = xfs_data_valid(inode, XFS_DATA_R);
    if (error == 0) {
#ifndef LINUX2_1
	struct inode *t = DATA_FROM_VNODE(inode);
	t->i_count++;
	
	while (file->f_pos < t->i_size && error >= 0) {
	  XFSDEB(XDEBREADDIR, ("xfs_readdir file->f_pos: %d t->i_size: %d\n",(int) file->f_pos,(int) t->i_size));
	    if (t->i_op &&
		t->i_op->default_file_ops &&
		t->i_op->default_file_ops->read) {
		buf = xfs_alloc(XFS_DIRENT_BLOCKSIZE);
		begin_offset = file->f_pos &~ (XFS_DIRENT_BLOCKSIZE - 1);
		offset = file->f_pos & (XFS_DIRENT_BLOCKSIZE - 1);
		
		XFSDEB(XDEBREADDIR, ("xfs_readdir begin_offset: %d offset: %d\n",(int)begin_offset,(int)offset));
		oldfs = get_fs();
		set_fs(KERNEL_DS);
		file->f_pos = begin_offset;
		file->f_reada = 0;
		error = t->i_op->default_file_ops->read(t, file, buf, XFS_DIRENT_BLOCKSIZE);
		set_fs(oldfs);
		if (error == XFS_DIRENT_BLOCKSIZE) {
		  XFSDEB(XDEBREADDIR, ("xfs_readdir error: %d\n",error));
		    while (offset < XFS_DIRENT_BLOCKSIZE) {
			XFSDEB(XDEBREADDIR,
			       ("xfs_readdir offset: %d namlen: %d offset2: %d\n",
				(int) offset,
				(int) ((struct xfs_dirent *) (buf + offset))->d_namlen,
				(int) (offset+begin_offset)));
			xn_file = xfs_cache_lookup (VNODE_TO_XNODE(inode),
						    ((struct xfs_dirent *) (buf + offset))->d_name,
						    ((struct xfs_dirent *) (buf + offset))->d_namlen);
			fileno = xn_file == NULL ?
			    ((struct xfs_dirent *) (buf + offset))->d_fileno :
			    XNODE_TO_VNODE(xn_file)->i_ino;
			if ((filldir_error = filldir (dirent,
						      ((struct xfs_dirent *) (buf + offset))->d_name,
						      ((struct xfs_dirent *) (buf + offset))->d_namlen,
						      offset+begin_offset,
						      fileno)) < 0) {
			    XFSDEB(XDEBREADDIR, ("xfs_readdir filldir: %d\n", filldir_error));
			    file->f_pos = offset + begin_offset;
			    file->f_reada = 0;
			    iput(t);
			    xfs_free(buf);
			    return 0;
			}
			offset += ((struct xfs_dirent *) (buf + offset))->d_reclen;
		    }
		    xfs_free(buf);
		} else if (error > 0) {
		    printk("XFS Panic! expected %d bytes, got %d in xfs_readdir\n", XFS_DIRENT_BLOCKSIZE, error);
		    error = -EINVAL;
		}
	    } else {
		error = -EINVAL;
		printk("XFS Panic! read not defined in cache file system\n");
	    }
#else
	struct dentry *t = DATA_FROM_VNODE(inode);
	struct qstr sqstr;
	ino_t ino;

	XFSDEB(XDEBREADDIR, ("xfs_readdir: inode: %p data:%p\n",inode, t));

	while (file->f_pos < t->d_inode->i_size && error >= 0) {
	    XFSDEB(XDEBREADDIR, ("xfs_readdir file->f_pos: %d t->i_size: %d\n",(int) file->f_pos,(int) t->d_inode->i_size));
	    buf = xfs_alloc(XFS_DIRENT_BLOCKSIZE);
	    begin_offset = file->f_pos &~ (XFS_DIRENT_BLOCKSIZE - 1);
	    offset = file->f_pos & (XFS_DIRENT_BLOCKSIZE - 1);
	    
	    XFSDEB(XDEBREADDIR, ("xfs_readdir begin_offset: %d offset: %d\n",(int)begin_offset,(int)offset));
	    file->f_pos = begin_offset;
	    error = read_exec(t,
			      file->f_pos,
			      buf,
			      XFS_DIRENT_BLOCKSIZE,
			      1);
	    if (error < 0) {
		printk("XFS: cannot read directory\n");
		return error;
	    }
	    if (error == 0)
		return error;

	    file->f_pos += error;
	    
	    if (error == XFS_DIRENT_BLOCKSIZE) {
		XFSDEB(XDEBREADDIR, ("xfs_readdir error: %d\n",error));
		while (offset < XFS_DIRENT_BLOCKSIZE) {
		    XFSDEB(XDEBREADDIR,
			   ("xfs_readdir offset: %d namlen: %d offset2: %d\n",
			    (int) offset,
			    (int) ((struct xfs_dirent *) (buf + offset))->d_namlen,
			    (int) (offset+begin_offset)));
		    sqstr.name = ((struct xfs_dirent *) (buf + offset))->d_name;
		    sqstr.len = ((struct xfs_dirent *) (buf + offset))->d_namlen;
		    sqstr.hash = full_name_hash(sqstr.name, sqstr.len);
		    
		    ino = find_inode_number(file->f_dentry->d_parent,
					    &sqstr);

		    fileno = ino ? ino
			: ((struct xfs_dirent *) (buf + offset))->d_fileno;

		    if ((filldir_error = filldir (dirent,
						  sqstr.name,
						  sqstr.len,
						  offset+begin_offset,
						  fileno)) < 0) {
			XFSDEB(XDEBREADDIR, ("xfs_readdir filldir: %d\n",
					     filldir_error));
			file->f_pos = offset + begin_offset;
			xfs_free(buf);
			return 0;
		    }
		    offset += ((struct xfs_dirent *) (buf + offset))->d_reclen;
		}
		xfs_free(buf);
	    } else if (error > 0) {
		printk("XFS Panic! expected %d bytes, got %d in xfs_readdir\n",
		       XFS_DIRENT_BLOCKSIZE, error);
		error = -EINVAL;
	    }
#endif /* LINUX2_1 */

	}
#ifndef LINUX2_1
	iput(t);
#endif /* LINUX2_1 */
    }
    
    return error;
}

static int
#ifndef LINUX2_1
xfs_readlink(struct inode *inode, char *buffer, int buflen)
#else
xfs_readlink(struct dentry *dentry, char *buffer, int buflen)
#endif /* LINUX2_1 */
{
    int error = 0;
#ifndef LINUX2_1
    struct inode *t;
#else
    struct inode *inode = dentry->d_inode;
    struct dentry *t;
#endif /* LINUX2_1 */
    
    XFSDEB(XDEBVNOPS, ("xfs_readlink\n"));
    
    error = xfs_data_valid(inode, XFS_DATA_R);
    if (error == 0)
    {
	off_t size;
	
	t = DATA_FROM_VNODE(inode);
	if (t == NULL)
	    return -EINVAL;
	size = (buflen > inode->i_size) ?
	    (inode->i_size) :
	    (buflen);
	error = read_exec(t, 0, buffer, size, 1);
    }
    
    return error;
}

#ifndef LINUX2_1
static int xfs_follow_link(struct inode * dir, struct inode * inode,
			   int flag, int mode, struct inode ** res_inode)
#else
static struct dentry *
xfs_follow_link(struct dentry *dentry, struct dentry *base)
#endif /* LINUX2_1 */
{
#ifndef LINUX2_1
    int error;
#else
    struct inode *inode = dentry->d_inode;
#endif /* LINUX2_1 */
    char *buffer;
#ifdef LINUX2_1
    int error;
#endif /* LINUX2_1 */

#ifndef LINUX2_1
    *res_inode = NULL;
    if (dir == NULL) {
	dir = current->fs->root;
	dir->i_count++;
    }
    if (!inode) {
	iput(dir);
	return -ENOENT;
    }
    if (!S_ISLNK(inode->i_mode)) {
	iput(dir);
	*res_inode = inode;
	return 0;
    }
    if (current->link_count > 5) {
	iput(dir);
	iput(inode);
	return -ELOOP;
    }
#endif /* LINUX2_1 */
    buffer = xfs_alloc(MAXPATHLEN + 1);
    if (buffer == NULL) {
#ifndef LINUX2_1
	iput(dir);
	iput(inode);
	return -ENOMEM;
#else
	dput(base);
	return NULL;
#endif /* LINUX2_1 */
    }
#ifndef LINUX2_1
    error = xfs_readlink(inode, buffer, MAXPATHLEN);
#else
    error = xfs_readlink(dentry, buffer, MAXPATHLEN);
#endif /* LINUX2_1 */
    if (error < 0) {
#ifndef LINUX2_1
	iput(dir);
	iput(inode);
#else
	dput(base);
#endif /* LINUX2_1 */
	xfs_free(buffer);
#ifndef LINUX2_1
	return error;
#else
	return NULL;
#endif /* LINUX2_1 */
    }
    if (error <= MAXPATHLEN)
	buffer[error]='\0';
#ifndef LINUX2_1
    current->link_count++;
    error = open_namei (buffer, 0, 0, res_inode, dir);
    current->link_count--;
    iput(inode);
    return error;
#else
    base = lookup_dentry(buffer, base, 1);
    xfs_free(buffer);
    return base;
#endif /* LINUX2_1 */
}


#ifndef LINUX2_1
static void
#else
static int
#endif /* LINUX2_1 */
xfs_release_file (struct inode *inode, struct file *file)
{
    struct xfs *xfsp = XFS_FROM_VNODE(inode);
    struct xfs_node *xn = VNODE_TO_XNODE(inode);
    int error = 0;
    
    XFSDEB(XDEBVNOPS, ("xfs_close\n"));
    
    if (file->f_mode & FMODE_WRITE && xn->flags & XFS_DATA_DIRTY)
	error = do_fsync(xfsp, xn, XFS_WRITE);
    
#ifndef LINUX2_1
    if (error)
	XFSDEB(XDEBVNOPS, ("xfs_release_file error: %d\n",error));
    
    return;
#else
    return error;
#endif /* LINUX2_1 */
}
